do -- Section [1]: Loading process 
	FlexBar2 = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceDB-2.0", "AceModuleCore-2.0", "AceDebug-2.0", "AceEvent-2.0")
	FlexBar2.OO = AceLibrary("AceOO-2.0");
	FlexBar2.Waterfall = AceLibrary("Waterfall-1.0");

	FlexBar2:RegisterDB("FlexBar2_DB");
	FlexBar2:RegisterDefaults('profile', {
		DataVersion = 1, -- Changed when incompatibilities occur in SavedVars during versions
		Buttons = {},
        Groups = {},
	});

	function FlexBar2:OnInitialize()
	    self:SetDebugging(false);
		self.ButtonList = {};
		self.ButtonMixinList = {};
		self.GroupMixinList = {};
		self.ButtonActivatorList = {};
		self.ButtonDeactivatorList = {};
		self.GroupActivatorList = {};
		self.ButtonWaterfallList = {};
		self.GroupWaterfallList = {};
		self.ButtonSettingLoaderList = {};
		self.ButtonDefaulters = {};
		self.Buttons = {};
        self.Groups = {};
		self:Debug("Searching for modules");
	    for _, Name in ipairs(self.ModuleList) do
            local Module = self.modules[Name];
			self:Debug("     Found Module: " .. Name);
			if(type(Module.ButtonMixin) == "table") then
				table.insert(self.ButtonMixinList, Module.ButtonMixin);
				self:Debug("          Found ButtonMixin");
				if(type(Module.ButtonMixin.Activate) == "function") then
					self:Debug("               Found Activator");			
					table.insert(FlexBar2.ButtonActivatorList, Module.ButtonMixin.Activate);
				end
				if(type(Module.ButtonMixin.Deactivate) == "function") then
					self:Debug("               Found Deactivator");
					table.insert(FlexBar2.ButtonDeactivatorList, Module.ButtonMixin.Deactivate);
				end
				if(type(Module.ButtonMixin.LoadSettings) == "function") then
					self:Debug("               Found Setting Loader");			
					table.insert(FlexBar2.ButtonSettingLoaderList, Module.ButtonMixin.LoadSettings);
				end
				if(type(Module.ButtonMixin.SetDefaults) == "function") then
					self:Debug("               Found Default Settings");			
					table.insert(FlexBar2.ButtonDefaulters, Module.ButtonMixin.SetDefaults);
				end
			end
			if(type(Module.GroupMixin) == "table") then
				self:Debug("          Found GroupMixin: ");
				self.GroupMixinList[Name] = Module.GroupMixin;
				if(type(Module.GroupMixin.Activate) == "function") then
					self:Debug("               Found Activator");
					table.insert(self.GroupActivatorList, Module.GroupMixin.Activate);
				end
			end
			if(type(Module.ButtonWaterfall) == "function") then
				self:Debug("          Found button waterfall");
				self.ButtonWaterfallList[Name] = Module.ButtonWaterfall;
			end
			if(type(Module.GroupWaterfall) == "function") then
				self:Debug("          Found group waterfall");
				self.GroupWaterfallList[Name] = Module.GroupWaterfall;
			end
		end
		
		self.ButtonSuperClass = self.OO.Class(unpack(self.ButtonMixinList));
			function self.ButtonSuperClass.prototype:init(Name, ...)
				FlexBar2.ButtonSuperClass.super.prototype.init(self);
				FlexBar2:Debug("          Initializing");
				if(type(Name) ~= "string") then
					FlexBar2:Debug("               Failed, invalid name");
					return false;
				end
				self.InfoTable = FlexBar2.db.profile.Buttons[Name];
				FlexBar2.Buttons[Name] = self;
                self.Name = Name;
				table.insert(FlexBar2.ButtonList, Name);
			end
			function self.ButtonSuperClass.prototype:LoadSettings()
				for _, Func in ipairs(FlexBar2.ButtonSettingLoaderList) do
					Func(self);
				end
			end
			function self.ButtonSuperClass.prototype:SetDefaults()
				for _, Func in ipairs(FlexBar2.ButtonDefaulters) do
					Func(self);
				end
			end
		self.GroupSuperClass = self.OO.Class(unpack(self.GroupMixinList));
			function self.GroupSuperClass.prototype:init(Name, ButtonTable, ...)
				FlexBar2:Debug("          Initializing");
				FlexBar2.GroupSuperClass.super.prototype.init(self);
                FlexBar2.Groups[Name] = self;
                self.MemberList = {};
				self.Name = Name;
                self.Count = #ButtonTable;
				self.InfoTable = FlexBar2.db.profile.Groups[Name];
                local MemberList = self.MemberList;
                local ButtonList = FlexBar2.Buttons;
                for _, ButtonName in ipairs(ButtonTable) do
                    ButtonList[ButtonName].Group = self;
                    MemberList[ButtonName] = ButtonList[ButtonName];
                    ButtonList[ButtonName].GroupMemberList = MemberList;
                    
                end
                
				for _, Func in ipairs(FlexBar2.GroupActivatorList) do
					Func(self, ...);
				end
			end
	end

	function FlexBar2:OnEnable()
        self:Debug("Enabling modules");
        for _, Module in ipairs(FlexBar2.ModuleList) do
            local ModuleRef = FlexBar2:GetModule(Module);
            if(type(ModuleRef.Activate) == "function") then
                self:Debug("   Enabling " .. Module);
                ModuleRef:Activate();
            end
        end
		self:SetupSlashCommands();
		self:SetupWaterfall();
		self:Debug("Generating buttons");
		for Name, InfoTable in pairs(self.db.profile.Buttons) do
			if(InfoTable.Generate == true) then
				self:Debug("     Generating " .. Name);
				self:Generate(Name);
                self:Debug("Generated");
			end
		end
        for Name, MemberList in pairs(self.db.profile.Groups) do
            FlexBar2:CreateGroup(Name, MemberList);
        end
	end

	function FlexBar2:OnDisable()
	    
	end

    FlexBar2.ModuleList = {};
	function FlexBar2:OnModuleCreated(Name, Module)
		table.insert(self.ModuleList, Name);
	end
end

do -- Section [2]: Configuration: GUI & slashcommands 
	-- Slashcommand handler: thanks th Mairelon for the inspiration on the API (though most code was written myself except for ToTable)
	FlexBar2.SlashCommands = {};
	local SlashCommands = FlexBar2.SlashCommands;
	SlashCommands.CommandList = {};
    SlashCommands.GroupList = {};
	function SlashCommands.Dispatch(Command)
		if(InCombatLockdown()) then FlexBar2:Print("Configuration is not possible in combat"); return; end
		if(type(Command) ~= "string") then FlexBar2:Print("Error: Command error. Please report"); return false; end
		local _, _, Function, Arguments = Command:find("^(%w+)[ ]*(.*)");
		if(Function == "help") then SlashCommands:Help(Arguments); return; end
		if(type(SlashCommands.CommandList[Function]) == "table" and type(SlashCommands.CommandList[Function][1]) == "function") then
			local Params = SlashCommands:ToTable(Arguments or "");
			if(SlashCommands:Verify(Function, Params, true) == false) then return false; end
			FlexBar2:Debug(Params);
			SlashCommands.CommandList[Function][1](Params);
		else
			SlashCommands:Help();
		end
	end
	
	FlexBar2:RegisterChatCommand({"/fb", "/flexbar", "/fb2", "/flexbar2"},
	{
	    type = 'text',
	    name = "FlexBar2 Slashcommands",
	    desc = "Flexbar2 slashcommand handler",
        usage = "gui will show up the config GUI. /fb help will tell you more about how to use slashcommands.",
	    get = false,
	    set = SlashCommands.Dispatch,
	});
	
	function SlashCommands:Help(Group)
        FlexBar2:Debug("Group:", Group)
        if(type(Group) == "string" and Group ~= "") then
            FlexBar2:Print("Available commands: ");
            for Command, InfoTable in pairs(SlashCommands.CommandList) do
                if(Group == InfoTable[2]) then
                    if(InfoTable[3]) then
                        FlexBar2:Print(InfoTable[3]);
                    else
                        FlexBar2:Print("/FlexBar2 " .. Command .. ": no help available");
                    end
                end
		    end
        elseif(Group == "" or type(Group) ~= "string") then
            FlexBar2:Print("Choose a group to get further help");
            for GroupName, _ in pairs(SlashCommands.GroupList) do
                FlexBar2:Print(GroupName);
            end
        end
	
		
	end

	function SlashCommands:ToTable(msg) -- nearly unmodified code from Mairelon's utility class used in FB1
		if msg == nil then return {} end
		-- pattern to return name=
		local pattern = "([%a_][%w_]*)="
		local index = 1
		local params = {}
		local firsti, lasti, capture = string.find(msg,pattern)
		-- while we have a name=, process the info after it.
		while capture do
			local varname=string.lower(capture)
			index = index+lasti
			firsti, lasti, capture = string.find(string.sub(msg, index),pattern)
			if not firsti then firsti=string.len(msg) else firsti=firsti-2 end
			local str = string.sub(msg,index, index+firsti)
			if(string.find(str, "[ ]+$")) then -- code to strip out spaces at the end
				str = string.sub(str, 1, (string.find(str, "()[ ]+$")) - 1);
			end
			-- str contain the rest of the equality ( varname = str )
			params[varname] = str;
		end
		return params;
	end

	function SlashCommands:Verify(Function, Params, Output)
		if(type(SlashCommands.CommandList[Function]) ~= "table"
			or type(SlashCommands.CommandList[Function][1]) ~= "function"
			or type(SlashCommands.CommandList[Function][4]) ~= "table") then
			if(Output) then FlexBar2:Print("Slashcommand not set up correctly, cancelling"); end
			return false;
		end
        if(type(Params["button"]) ~= "string" and type(Params["group"]) == "string") then
            Params.button = FlexBar2:GroupNameToButtonRange(Params["group"]);
        end
		for Param, Info in pairs(SlashCommands.CommandList[Function][4]) do 
			if(type(Params[Param]) ~= "string") then
				if(Info.Required == true) then
					if(Output) then FlexBar2:Print("Required parameter [" .. Param .. "] not set, cancelling"); end
					return false;
				elseif(type(Info.DefaultValue) ~= "string") then
					break;
				end
			end

			if(type(Info.DefaultValue) == "string" and type(Params[Param]) ~= "string") then
				Params[Param] = Info.DefaultValue;
			end

			if(type(Info.AllowedValues) == "table") then
				if(not FlexBar2:ValueIsInTable(Info.AllowedValues, Params[Param])) then
					if(Output) then FlexBar2:Print("[" .. Params[Param] .. "] is not in the list of allowed values for " .. Param); end
					return false;
				end
			end
            if(type(Info.VerifyFunc) == "function") then
                local VerifyReturn = Info.VerifyFunc(Params[Param]);
                if(VerifyReturn ~= true) then
				    if(Output) then FlexBar2:Print("Verification failed for [" .. Param .. "]: " .. (VerifyReturn or ("], Invalid value [" .. Params[Param] .. "]"))); end
				    return false;
			    end
            end             
		end
		return true;
	end

	--[[
		Command: The name of the command we are adding
		Function: The command that this will be passed to
		Group: What group the function is in (for the help box)
		Help: function usage which will be displayed in the /FlexBar2 help <command> message
	--]]
	function SlashCommands:AddCommand(Command, Function, Group, Help)
		if(type(Command) ~= "string" or type(Function) ~= "function" or SlashCommands.CommandList[Command] == "table") then
			return false;
		else
			Command = Command:lower();
		end
		if(type(Group) ~= "string") then Group = "main"; end
        SlashCommands.GroupList[Group] = true;
		SlashCommands.CommandList[Command] = {Function, Group, Help, {}};
	end
	
	--[[
		Command: The name of the command we are adding parameters to
		Param: The name of the parameter we are adding
		Required: wether to error out if the parameter is not set
		Types: what type the parameter can be (string or table [ranges are seen as tables])
		AllowedValues: what values are allowed for this parameter (example: {"true", "false"})
		Default: what default value the parameter takes if it is ommited
	--]]
	function SlashCommands:AddParam(Command, Param, Required, AllowedValues, DefaultValue, VerifyFunc) 
		if(type(Command) ~= "string"
			or type(Param) ~= "string"
            or type(SlashCommands.CommandList[Command]) ~= "table"
			or type(SlashCommands.CommandList[Command][4][Param]) == "table") then
				return false;
		end
		SlashCommands.CommandList[Command][4][Param] =
		{
			Required = Required,
			AllowedValues = AllowedValues,
			DefaultValue = DefaultValue,
			VerifyFunc = VerifyFunc,
		};
	end
	
    function SlashCommands.VerifyExistance(Button)
        if(not FlexBar2:ButtonsExist(Button)) then
            return "One or more buttons do not exist";
        else
            return true;
        end
    end

    function SlashCommands.VerifyNonExistance(Button)
        if(FlexBar2:ButtonsExist(Button)) then
            return "One or more buttons already exist";
        else
            return true;
        end
    end

    function SlashCommands.VerifyNotGrouped(Button)
        local Existance = SlashCommands.VerifyExistance(Button);
        if(Existance ~= true) then
            return Existance;
        else
            if(not FlexBar2:ButtonIsGrouped(Button)) then
                return true;
            else
                return "One or more buttons are already grouped";
            end
        end
    end
    
    function SlashCommands.VerifyGrouped(Button)
        local Existance = SlashCommands.VerifyExistance(Button);
        if(Existance ~= true) then
            return Existance;
        else
            if(FlexBar2:ButtonIsGrouped(Button)) then
                return true;
            else
                return "One or more buttons are not grouped";
            end
        end
    end
    
	local function SlashCommand_Create(Params)
		for _, ButtonName in ipairs(FlexBar2:ToButtonTable(Params.button)) do
			FlexBar2:Create(ButtonName);
		end
	end
	
	local function SlashCommand_Delete(Params)
		for _, ButtonName in ipairs(FlexBar2:ToButtonTable(Params.button)) do
			FlexBar2:Delete(ButtonName);
		end
	end

    local function SlashCommand_List()
        FlexBar2:Print("List of buttons: ");
        for ButtonName, _ in pairs(FlexBar2.Buttons) do
            FlexBar2:Print("[" .. ButtonName .. "]");
        end
    end
	
    local function SlashCommand_Group(Params)
        FlexBar2:CreateGroup(Params.name, FlexBar2:ToButtonTable(Params.button));
    end

    local function SlashCommand_Ungroup(Params)
        local ButtonTable = FlexBar2:ToButtonTable(Params.button);
        for _, ButtonName in ipairs(ButtonTable) do
            FlexBar2:RemoveButtonFromGroup(ButtonName);
        end
    end

	function FlexBar2:SetupSlashCommands()
        local VerifyExistance = SlashCommands.VerifyExistance;
        local VerifyNonExistance = SlashCommands.VerifyNonExistance;
        local VerifyGrouped = SlashCommands.VerifyGrouped;
        local VerifyNotGrouped = SlashCommands.VerifyNotGrouped;
		SlashCommands:AddCommand("gui", function() FlexBar2.Waterfall:Open("FlexBar2"); end, "main", "/FlexBar2 gui");
		SlashCommands:AddCommand("create", SlashCommand_Create, "main", "/FlexBar2 create button=<Buttons>");
		    SlashCommands:AddParam("create", "button", true, nil, nil, VerifyNonExistance);
		SlashCommands:AddCommand("delete", SlashCommand_Delete, "main", "/FlexBar2 delete button=<Buttons>");
			SlashCommands:AddParam("delete", "button", true, nil, nil, VerifyExistance);
        SlashCommands:AddCommand("list", SlashCommand_List, "main", "/FlexBar2 list");
        SlashCommands:AddCommand("group", SlashCommand_Group, "main", "/FlexBar2 group button=<Buttons> name=<Group name>");
            SlashCommands:AddParam("group", "button", true, nil, nil, VerifyNotGrouped);
            SlashCommands:AddParam("group", "name", true, nil, nil, function(Name) if(string.find(Name, "^[%w_]+$")) then return true; else return "Invalid group name"; end end);
        SlashCommands:AddCommand("ungroup", SlashCommand_Ungroup, "main", "/FlexBar2 ungroup group=<Group>");
			SlashCommands:AddParam("ungroup", "button", true, nil, nil, VerifyGrouped);
	end
	
	function FlexBar2:SetupWaterfall() -- Called in OnEnable
		local Waterfall = self.Waterfall;
		self:Debug("Setting up Waterfall");
		local WaterfallTree = 
		{
			{
				text = "General Settings",
				id = "gnrlset",
				isOpen = true,
			},
			{
				text = "Button Control",
				id = "buttonctrl",
				isOpen = false,
			},
			{
				text = "Group Control",
				id = "groupctrl",
				isOpen = false,
			},
		};
		local buttonctrl_CreateTemp, buttonctrl_EditSave, buttonctrl_EditTemp;
		local ButtonTable;
        local groupctrl_EditSetTemp, groupctrl_EditSetSave, groupctrl_ButtonCheckOverride, groupctrl_CreateGroupNameTemp, groupctrl_CreateButtonsTemp, groupctrl_RemoveButtonTemp, groupctrl_DisbandGroupTemp;
		local function BuildWaterfallTree(Id)
			FlexBar2:Debug("GUI Id: ", Id);
			if not Id then return ""; end
			if(Id == "gnrlset") then
			elseif(Id == "buttonctrl") then
                FlexBar2:Debug("editsave: ", buttonctrl_EditSave);
				if(buttonctrl_EditSave ~= nil and buttonctrl_EditSave ~= "" and groupctrl_ButtonCheckOverride ~= true) then
					ButtonTable = FlexBar2:ToButtonTable(buttonctrl_EditSave);
					FlexBar2:Debug("ButtonTable: ", #ButtonTable);
					if(#ButtonTable == "0") then buttonctrl_EditSave = nil; end
					for _, ButtonName in ipairs(ButtonTable) do
						if(type(FlexBar2.Buttons[ButtonName]) ~= "table") then
							FlexBar2:Print("non-existant button(s) have been inputted");
							buttonctrl_EditSave = nil;
                            ButtonTable = nil;
						end
					end
				end
				if(type(ButtonTable) ~= "table") then
					FlexBar2:Debug("Waterfall: ButtonCtrl main");
					-- Create Button --
					Waterfall:AddControl(
					"type", "label",
					"text", "Create Button: ");
					
					Waterfall:AddControl(
					"type","textbox",
					"width",200,
					"fullRefresh", true,
					"noNewLine", true,
					"setFunc", function()
                        local ButtonTable = FlexBar2:ToButtonTable(buttonctrl_CreateTemp);
                            if(type(ButtonTable) == "table") then
                            buttonctrl_EditSave = buttonctrl_CreateTemp;
                            for _, ButtonName in ipairs(ButtonTable) do
                                FlexBar2:Create(ButtonName);
                            end
                        end
                    end,
					"getFunc", function() return ""; end,
					"changedFunc", function(ButtonName) buttonctrl_CreateTemp = ButtonName; end);
					
					Waterfall:AddControl(
					"type","button",
					"text","Create",
					"fullRefresh", true,
					"noNewLine",true,
					"width",80,
					"causeRefresh",true,
					"execFunc", function()
                        local ButtonTable = FlexBar2:ToButtonTable(buttonctrl_CreateTemp);
                            if(type(ButtonTable) == "table") then
                            buttonctrl_EditSave = buttonctrl_CreateTemp;
                            for _, ButtonName in ipairs(ButtonTable) do
                                FlexBar2:Create(ButtonName);
                            end
                        end
                    end);
					
					-- Edit Button --
					Waterfall:AddControl(
					"type", "label",
					"text", "Edit Button: ");
					
					Waterfall:AddControl(
					"type","textbox",
					"width",200,
					"noNewLine", true,
					"fullRefresh", true,
					"setFunc", function(ButtonName) buttonctrl_EditSave = ButtonName end,
					"changedFunc", function(ButtonName) buttonctrl_EditTemp = ButtonName end);
					
					Waterfall:AddControl(
					"type","button",
					"text","Edit",
					"noNewLine",true,
					"fullRefresh", true,
					"width",80,
					"execFunc", function() buttonctrl_EditSave = buttonctrl_EditTemp end);
					return "Button control";
				else
					FlexBar2:Debug("Waterfall: Buttonctrl edit");
					for Module, Data in pairs(FlexBar2.ButtonWaterfallList) do
						if(type(Data) == "function") then
							Data("edit", buttonctrl_EditSave, ButtonTable);
						end
					end
					-- Delete Button --
					Waterfall:AddControl(
					"type", "button",
					"text","Delete Button",
					"fullRefresh", true,
					"width", 100,
					"execFunc", function()
                        for _, ButtonName in ipairs(ButtonTable) do
                            FlexBar2:Delete(ButtonName);
                        end
                        buttonctrl_EditSave = nil;
                        ButtonTable = nil;
                        if(groupctrl_ButtonCheckOverride == true) then
                            Waterfall:Open("FlexBar2", "groupctrl");
                            groupctrl_ButtonCheckOverride = nil;
                        end
                        end);
					
					-- Cancel Button --
					Waterfall:AddControl(
					"type", "button",
					"text","Cancel changes",
					"fullRefresh", true,
					"noNewLine", true,
					"width", 130,
					"execFunc", function()
						buttonctrl_EditSave = nil;
                        ButtonTable = nil;
                        if(groupctrl_ButtonCheckOverride == true) then
                            Waterfall:Open("FlexBar2", "groupctrl");
                            groupctrl_ButtonCheckOverride = nil;
                        end
						end,
					"execArg1");

					-- Save Button --
					Waterfall:AddControl(
					"type", "button",
					"text","Save changes",
					"fullRefresh", true,
					"noNewLine", true,
					"width", 130,
					"execFunc", function()
						for LoopNumber, ButtonName in ipairs(ButtonTable) do
                            for Module, Data in pairs(FlexBar2.ButtonWaterfallList) do
                                if(type(Data) == "function") then
                                    Data("save", LoopNumber, ButtonName);
                                end
                            end
                        end
						
						buttonctrl_EditSave = nil;
                        ButtonTable = nil;
                        if(groupctrl_ButtonCheckOverride == true) then
                            Waterfall:Open("FlexBar2", "groupctrl");
                            groupctrl_ButtonCheckOverride = nil;
                        end
					end);

					return "Edit button settings";
				end				
			elseif(Id == "groupctrl") then
                if((type(groupctrl_EditSetSave) ~= "string" or type(FlexBar2.Groups[groupctrl_EditSetSave]) ~= "table")) then
                    local CreateGroup = function()
                        ButtonTable = FlexBar2:ToButtonTable(groupctrl_CreateButtonsTemp);
                        local GroupName = groupctrl_CreateGroupNameTemp;
                        if(type(ButtonTable) == "table" and string.find(GroupName, "^[%w_]+$")) then
                            FlexBar2:CreateGroup(GroupName, ButtonTable);
                            groupctrl_ButtonCheckOverride = true; 
                            Waterfall:Open("FlexBar2", "buttonctrl");
                        else
                            FlexBar2:Print("Invalid Group name and/or Button range");
                        end
                    end
                    
                    Waterfall:AddControl(
                    "type", "label",
                    "text", "Create Group: ");

                    Waterfall:AddControl(
                    "type","textbox",
                    "width",200,
                    "noNewLine", true,
                    "fullRefresh", true,
                    "setFunc", CreateGroup,
                    "changedFunc", function(GroupName) groupctrl_CreateGroupNameTemp = GroupName end);

                    Waterfall:AddControl(
                    "type", "label",
                    "text", "                         Buttons: ");

                    Waterfall:AddControl(
                    "type","textbox",
                    "width",200,
                    "noNewLine", true,
                    "fullRefresh", true,
                    "setFunc", CreateGroup,
                    "changedFunc", function(ButtonRange) groupctrl_CreateButtonsTemp = ButtonRange end);
                    
                    Waterfall:AddControl(
                    "type","button",
                    "text","Create",
                    "noNewLine",true,
                    "fullRefresh", true,
                    "width",80,
                    "execFunc", CreateGroup);

                    Waterfall:AddControl(
                    "type", "label",
                    "text", "Disband group: ");

                    Waterfall:AddControl(
                    "type","textbox",
                    "width",200,
                    "noNewLine", true,
                    "fullRefresh", true,
                    "setFunc", function()
                        FlexBar2:DisbandGroup(groupctrl_DisbandGroupTemp);
                    end,
                    "changedFunc", function(GroupName) groupctrl_DisbandGroupTemp = GroupName end);

                    Waterfall:AddControl(
                    "type","button",
                    "text","Disband",
                    "noNewLine",true,
                    "fullRefresh", true,
                    "width",80,
                    "execFunc", function()
                        FlexBar2:DisbandGroup(groupctrl_DisbandGroupTemp);
                    end);

                    Waterfall:AddControl(
                    "type", "label",
                    "text", "Ungroup button(s): ");

                    Waterfall:AddControl(
                    "type","textbox",
                    "width",200,
                    "noNewLine", true,
                    "fullRefresh", true,
                    "setFunc", function()
                        local ButtonTable = FlexBar2:ToButtonTable(groupctrl_RemoveButtonTemp);
                        if(type(ButtonTable) == "table") then
                            for _, ButtonName in ipairs(ButtonTable) do
                                FlexBar2:RemoveButtonFromGroup(ButtonName);
                            end
                        end
                    end,
                    "changedFunc", function(ButtonRange) groupctrl_RemoveButtonTemp = ButtonRange end);
                    
                    Waterfall:AddControl(
                    "type","button",
                    "text","Remove",
                    "noNewLine",true,
                    "fullRefresh", true,
                    "width",80,
                    "execFunc", function()
                        local ButtonTable = FlexBar2:ToButtonTable(groupctrl_RemoveButtonTemp);
                        if(type(ButtonTable) == "table") then
                            for _, ButtonName in ipairs(ButtonTable) do
                                FlexBar2:RemoveButtonFromGroup(ButtonName);
                            end
                        end
                    end);

                    Waterfall:AddControl(
                    "type", "label",
                    "text", "Edit Group settings: ");
                    
                    Waterfall:AddControl(
                    "type","textbox",
                    "width",200,
                    "noNewLine", true,
                    "fullRefresh", true,
                    "setFunc", function()
                        if(type(FlexBar2.Groups[groupctrl_EditSetTemp]) == "table") then
                            groupctrl_ButtonCheckOverride = true;
                            ButtonTable = FlexBar2:ToButtonTable(FlexBar2.Groups[groupctrl_EditSetTemp].MemberList);
                            Waterfall:Open("FlexBar2", "buttonctrl");
                        end
                    end,
                    "changedFunc", function(GroupName) groupctrl_EditSetTemp = GroupName end);
                    
                    Waterfall:AddControl(
                    "type","button",
                    "text","Edit",
                    "noNewLine",true,
                    "fullRefresh", true,
                    "width",80,
                    "execFunc", function()
                        if(type(FlexBar2.Groups[groupctrl_EditSetTemp]) == "table") then
                            groupctrl_ButtonCheckOverride = true;
                            ButtonTable = FlexBar2:ToButtonTable(FlexBar2.Groups[groupctrl_EditSetTemp].MemberList);
                            Waterfall:Open("FlexBar2", "buttonctrl");
                        end
                    end);
                    return "Group control";
                end
			end
		end
		Waterfall:Register('FlexBar2',
		'title', 'FlexBar2 Config',
		'width', 800,
		'height', 500,
		'tree', WaterfallTree,
		'children', BuildWaterfallTree);
	end
end

do -- Section [3]: Button control 
	function FlexBar2:Create(Name)
		if(type(FlexBar2.db.profile.Buttons[Name]) ~= "table") then
			FlexBar2.db.profile.Buttons[Name] = { Generate = true};
			FlexBar2:Generate(Name);
			return true;
		else
			FlexBar2:Print("Button already exists");
			return false;
		end
	end

	function FlexBar2:Delete(Name)
		if(type(FlexBar2.db.profile.Buttons[Name]) == "table") then
			FlexBar2.db.profile.Buttons[Name] = nil;
			local Button = FlexBar2.Buttons[Name];
			for _, Func in ipairs(FlexBar2.ButtonDeactivatorList) do
				Func(Button);
			end
            FlexBar2:RemoveButtonFromGroup(Name);
			FlexBar2.Buttons[Name] = nil;
			return true;
		else
			FlexBar2:Print("Cannot delete non-existant button");
			return false;
		end
		
	end

	function FlexBar2:Generate(Name)
		if(self.Buttons[Name] ~= nil) then
			self:Debug("          Error: Button already exists, aborting");
			return false;
		end
		self.ButtonSuperClass:new(Name);
        local ButtonList = FlexBar2.Buttons;
        if(type(self.db.profile.Buttons[Name].Modules) ~= "table") then self.db.profile.Buttons[Name].Modules = {}; end
        local ModuleList = self.db.profile.Buttons[Name].Modules;
        for _, ModuleName in ipairs(FlexBar2.ModuleList) do
            local Module = FlexBar2:GetModule(ModuleName);
            local ModuleVersion = FlexBar2:GetModule(ModuleName).Version;
            local ButtonMixin = Module.ButtonMixin;
            if(type(ButtonMixin) == "table") then
                if(type(ButtonMixin.Initialize) == "function") then ButtonMixin.Initialize(ButtonList[Name]); end
                if((type(ModuleList) ~= "table" or ModuleList[ModuleName] ~= ModuleVersion) and type(ButtonMixin.SetDefaults) == "function") then
                    ButtonMixin.SetDefaults(ButtonList[Name]);
                end
                if(type(ButtonMixin.LoadSettings) == "function") then ButtonMixin.LoadSettings(ButtonList[Name]); end
                if(type(ButtonMixin.Activate) == "function") then ButtonMixin.Activate(ButtonList[Name]); end
                ModuleList[ModuleName] = ModuleVersion;
            end
        end
        return true;
	end
end

do -- Section [4]: Group control
    function FlexBar2:CreateGroup(Name, ButtonTable)
        self.GroupSuperClass:new(Name, ButtonTable);
        self.db.profile.Groups[Name] = ButtonTable;
    end

	function FlexBar2:DisbandGroup(Name)
        if(type(FlexBar2.Groups[Name]) == "table") then
            local Group = FlexBar2.Groups[Name];
            for ButtonName, Button in pairs(Group.MemberList) do
                Button.Group = nil;
                Button.GroupMemberList = nil;
            end
            FlexBar2.Groups[Name] = nil;
            FlexBar2.db.profile.Groups[Name] = nil;
        end
    end

    function FlexBar2:AddButtonToGroup(Group, Name)
		if(not FlexBar2:ButtonIsGrouped(Name)) then
            FlexBar2.Groups[Group].MemberList[Name] = FlexBar2.Buttons[Name];
        end
	end

	function FlexBar2:RemoveButtonFromGroup(Name)
		local Button = FlexBar2.Buttons[Name];
		if(type(Button.GroupMemberList) == "table") then
			Button.GroupMemberList[Name] = nil;
            local Group = Button.Group;
            Group.Count = Group.Count - 1;
            FlexBar2:RemoveByValue(Button.Group.InfoTable, Name);
            Button.Group = nil;
            Button.GroupMemberList = nil;
            if(Group.Count == 0) then
                FlexBar2:DisbandGroup(Group.Name);
            end
		end
	end
end

do -- Section [5]: Utility
	function FlexBar2:GetSetting(ButtonTable, Setting)
		if(type(ButtonTable) ~= "table" or type(Setting) ~= "string") then return nil; end
		return FlexBar2.db.profile.Buttons[ButtonTable[1]][Setting];
	end

	function FlexBar2:ApplySetting(ButtonTable, Setting, Value)
		if(type(ButtonTable) == "table") then
			for _, ButtonName in ipairs(ButtonTable) do
				FlexBar2.db.profile.Buttons[ButtonName][Setting] = Value;
			end
			return true;
		elseif(type(ButtonTable) == "string") then
			FlexBar2.db.profile.Buttons[ButtonTable][Setting] = Value;
			return true;
		else
			return false;
		end
	end
	
	function FlexBar2:ToButtonTable(ButtonRange)
		if(type(ButtonRange) == "string") then
			local ButtonTable = {};
            if(ButtonRange:find("%d+-%d+")) then
				local _, _, First, Last = ButtonRange:find("^(%d+)-(%d+)$");
				for ButtonNumber = First, Last do
					table.insert(ButtonTable, tostring(ButtonNumber));
				end
			elseif(ButtonRange:find("[%w_];[%w_]")) then
				ButtonTable = { string.split(";", ButtonRange) };
			else
				ButtonTable = { ButtonRange };
			end
			for _, ButtonName in ipairs(ButtonTable) do
				if(not ButtonName:find("^[%w_]+$")) then
					return false;
				end
			end
			return ButtonTable;
		elseif(type(ButtonRange) == "table") then
            local ButtonTable = {};
			for Name, Object in pairs(ButtonRange) do
                table.insert(ButtonTable, Name);
            end
            if(#ButtonTable == 0) then
                return ButtonRange;
            else
                return ButtonTable;
            end
		else
			return false;
		end		
	end

    function FlexBar2:GroupNameToButtonRange(GroupName)
        if(type(FlexBar2.Groups[GroupName]) == "table" and type(FlexBar2.Groups[GroupName].MemberList) == "table") then
            local ButtonTable = FlexBar2:ToButtonTable(FlexBar2.Groups[GroupName].MemberList);
            if(type(ButtonTable) == "table") then
                return table.concat(ButtonTable, ";");
            end
        end
        return false;
    end
	
	function FlexBar2:ButtonsExist(Button)
		if(type(Button) == "string") then
            Button = FlexBar2:ToButtonTable(Button);
		end
		if(type(Button) == "table") then
            local ButtonList = FlexBar2.Buttons;
			for _, ButtonName in ipairs(Button) do
				if(ButtonList[ButtonName] == nil) then return false; end
			end
			return true;
		else
			return false;
		end
	end

    function FlexBar2:ButtonIsGrouped(Button)
		if(type(Button) == "string") then
			Button = FlexBar2:ToButtonTable(Button);
		end
		if(type(Button) == "table") then
            local ButtonList = FlexBar2.Buttons;
			for _, ButtonName in ipairs(Button) do
                if(type(ButtonList[ButtonName].Group) == "table") then return true; end
			end
			return false;
		else
			return false;
		end
	end
	
	function FlexBar2:ValueIsInTable(Table, Value)
		for n, v in pairs(Table) do
            if(v == Value) then return true; end
        end
        return false;
	end

    function FlexBar2:RemoveByValue(tbl, val)
        if(type(tbl) ~= "table") then return false; end
        for n, v in pairs(tbl) do
            if(v == val) then
                table.remove(tbl, n);
                return true;
            end
        end
        return false;
    end

    function FlexBar2:GetKeyByValue(tbl, val)
        for n, v in pairs(tbl) do
            if(v == val) then
                return n;
            end
        end
        return false;
    end
end
